/************************************************************************
 *                                                                      *
 *    Copyright (c) 1991 Wilhelm Valder, Regine Freitag, Wolfgang Dzida *
 *                        All rights reserved                           *
 *                                                                      *
 *                                                                      *
 *    Permission to copy this program or any part of this program is    *
 *    granted in case of personal, non-commercial use only. Any use     *
 *    for profit or other commercial gain without written permissions   *
 *    of the authors is prohibited.                                     *
 *                                                                      *
 *    You may freely distribute this program to friends, colleagues,    *
 *    etc.  As a courtesy we would appreciate it if you left our names  *
 *    on the code as the authors.                                       *
 *    We also, of course, disavow any liabilities                       *
 *    for any bizarre things you may choose to do with the system.      *
 *                                                                      *
 *    We would appreciate though your letting us know about any         *
 *    extensions, modifications, project suggestions, etc.              *
 *    Please mail all bugs, comments, complaints and suggestions to:    *
 *                                                                      *
 *                                                                      *
 *    GMD (German National Research Center for Computer Science)        *
 *    Institut fuer Systemtechnik (F2G2)                                *
 *    P.O. Box 1240                                                     *
 *    Schloss Birlinghoven                                              *
 *    D-5205 St. Augustin 1, West Germany                               *
 *                                                                      *
 *    email:    dzida@kmx.gmd.dbp.de                                    *
 *              freitag@gmdzi.uunet                                     *
 *                                                                      *
 ************************************************************************/
#include "fm.h"

static char SccsID[] = "@(#)sort.c	1.2\t1/27/93";            /* sccs keyword */

static int compareByName();
static int compareByAccess();
static int compareByMod();
static int compareBySize();

static int (* sortFunctions[])() = {
  (int (*) ()) compareByName,
  (int (*) ()) compareByAccess,
  (int (*) ()) compareByMod,
  (int (*) ()) compareBySize
};

sortItemList(w)
     dirWindow *w;
{
  dirItem *listSort();

  if (w->sortFlag > 0)
    w->itemList = listSort(w->itemList, sortFunctions[w->sortFlag - 1]);
}

static int compareByName(olditem,newitem)
     dirItem *olditem, *newitem;
{
  return((strcmp(ITEM_NAME(olditem),ITEM_NAME(newitem)) <= 0) ? 0 : 1);
}

static int compareByAccess(olditem, newitem)
     dirItem *olditem, *newitem;
{
  return((olditem->sbp->st_atime > newitem->sbp->st_atime) ? 0 : 1);
}

static int compareByMod(olditem, newitem)
     dirItem *olditem, *newitem;
{
  return((olditem->sbp->st_mtime > newitem->sbp->st_mtime) ? 0 : 1);
}

static int compareBySize(olditem, newitem)
     dirItem *olditem, *newitem;
{
  return((olditem->sbp->st_size > newitem->sbp->st_size) ? 0 : 1);
}

/*  listSort( list, compare_routine ) -> sorted_list

    Assumes a list of structures, with a pointer to "next" as the first
    field.  It reorders the list into ascending order, and returns the
    new first node's address.  It is order N log(N).

    The compare routine should return 0 if the items are in order, and 1
    if they are not.  If the compare routine returns 0 in case of
    equality, the sort will be stable.

    This routine depends upon compiler-dependant struct layout, but this
    assumption is likely to be fairly commonly valid.
*/

typedef dirItem *(list_pair_type[2]);

dirItem *listSort(list, compare)
    dirItem *list;
    int (*compare)();
{
    list_pair_type old, new; 	/* front of old stuff, back of new */
    int n;              	/* current new list */
    int o;              	/* current old list */
    long m, hm;         /*	    m is merge length, hm is half of that */
    long items_merged;  	/* counts how many items have been merged */
    long count[2];      	/* counts entries drawn from "old" lists */
    dirItem front[2];   	/* artificial front-of-list */

    front[0].next = NULL;
    front[1].next = list;

/*  The basic notion of this sort is to make sorted sublists longer and
    longer by merging.  On the Nth pass through the list, sorted
    sublists of length 2^(N-1) are produced.  Eventually, the entire
    list is sorted, in log2(N)+1 passes through the list.  There is
    extra bookkeeping overhead, but minimal extra storage space needed.
    Counts and clever pointer management substitute for extra "glue"
    nodes.

    while more than one list
       while not at end of composite lists
             for each merge_length(m) block
                   merge first items in lists onto current output list
             toggle current output list
*/

    m = 1;
    hm = 0;
    while( front[1].next ){
        /* log2(N)+1 times through (since m doubles every time) */
        new[0] = &front[0];
        new[1] = &front[1];
        old[0] = front[0].next;
        old[1] = front[1].next;
        count[0] = count[1] = 0;
        n = 0;
        for( items_merged=0; old[0] || old[1]; ++items_merged ){
            /* N times through (each time through consumes one item
               from one or the other of the old lists)
            */
            if( items_merged>=m ){
                items_merged = 0;
                n = 1-n;
                count[0] = count[1] = 0;
            }
            if( old[0] && old[1] && (count[0]<hm) && (count[1]<hm) ){
                /*  Compare the two items at the heads of the input
                    lists, and put the smaller on the current output
                    list
                */
                o = (*compare)( old[0], old[1] );
                new[n]->next = old[o];
                old[o] = old[o]->next;
                ++count[o];
            }
            else{
                /*  One of two input lists exhausted... either because
                    of end of a composite list, or because of end of
                    logical list (too many gotten from one branch or
                    another).
                */
                o = ((old[0] && (count[0]<hm)) ?0:1);
                new[n]->next = old[o];
                old[o] = old[o]->next;
            }

            new[n] = new[n]->next;
        }
        new[0]->next = new[1]->next = 0;

        hm = m;
        m *= 2;
    }
    return( front[0].next );

}
